"""Module that implements the User class."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from .const import UserAttribute as Attribute
from .entity import Entity

if TYPE_CHECKING:
    from .system import System

ADD_LOCK = f"{Attribute.LOCK_IDS}.1"


class User(Entity):
    """Describe a Vivint user."""

    def __init__(self, data: dict, system: System):
        """Initialize a user."""
        super().__init__(data)
        self._system = system

    def __repr__(self) -> str:
        """Return custom __repr__ of user."""
        return f"<{self.__class__.__name__} {self.id}, {self.name}{' (admin)' if self.is_admin else ''}>"

    @property
    def has_lock_pin(self) -> bool:
        """Return True if the user has pins."""
        return bool(self.data[Attribute.HAS_LOCK_PIN])

    @property
    def has_panel_pin(self) -> bool:
        """Return True if the user has pins."""
        return bool(self.data[Attribute.HAS_PANEL_PIN])

    @property
    def has_pins(self) -> bool:
        """Return True if the user has pins."""
        return bool(self.data[Attribute.HAS_PINS])

    @property
    def has_remote_access(self) -> bool:
        """Return True if the user has remote access."""
        return bool(self.data[Attribute.REMOTE_ACCESS])

    @property
    def id(self) -> int:  # pylint: disable=invalid-name
        """User's id."""
        return int(self.data[Attribute.ID])

    @property
    def is_admin(self) -> bool:
        """Return True if the user is an admin."""
        return bool(self.data[Attribute.ADMIN])

    @property
    def is_registered(self) -> bool:
        """Return True if the user is registered."""
        return bool(self.data[Attribute.REGISTERED])

    @property
    def lock_ids(self) -> list[int]:
        """User's lock ids."""
        return cast(list[int], self.data.get(Attribute.LOCK_IDS, []))

    @property
    def name(self) -> str:
        """User's name."""
        return str(self.data[Attribute.NAME])

    def handle_pubnub_message(self, message: dict) -> None:
        """Handle a pubnub message addressed to this user."""
        if ADD_LOCK in message:
            message[Attribute.LOCK_IDS] = self.lock_ids + [message[ADD_LOCK]]
            del message[ADD_LOCK]
        super().handle_pubnub_message(message)



"""Utilities module."""

from __future__ import annotations

import asyncio
import base64
import hashlib
import logging
import os
import re
from typing import Any, Callable, Coroutine, Iterable, TypeVar
from warnings import warn

_LOGGER = logging.getLogger(__name__)

_T = TypeVar("_T")


def first_or_none(lst: Iterable[_T], predicate: Callable[[_T], Any]) -> _T | None:
    """Return the first occurrence or `None` of an iterable, given a predicate."""
    filter_iter = filter(predicate, lst)
    return next(filter_iter, None)


def add_async_job(
    target: Callable | Coroutine, *args: Any
) -> asyncio.Task | asyncio.Future:
    """Add a callable to the event loop."""
    loop = asyncio.get_event_loop()
    task: asyncio.Future | asyncio.Task

    if asyncio.iscoroutine(target):
        task = loop.create_task(target)
    elif asyncio.iscoroutinefunction(target):
        task = loop.create_task(target(*args))
    else:
        task = loop.run_in_executor(None, target, *args)

    return task


def send_deprecation_warning(old_name: str, new_name: str) -> None:
    """Send a deprecation warning."""
    message = f"{old_name} has been deprecated in favor of {new_name}, the alias will be removed in the future"
    warn(
        message,
        DeprecationWarning,
        stacklevel=2,
    )
    _LOGGER.warning(message)


def generate_code_challenge() -> tuple[str, str]:
    """Generate PKCE code verifier and challenge for authentication."""
    code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
    code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)

    code_hash = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_hash).decode("utf-8")
    code_challenge = code_challenge.replace("=", "")

    return (code_verifier, code_challenge)


def generate_state() -> str:
    """Generate state."""
    return base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")



"""Module that implements the Vivint class."""

from __future__ import annotations

import asyncio
import logging

import aiohttp
from aiohttp.client_exceptions import ClientConnectionError
from pubnub.pnconfiguration import PNConfiguration
from pubnub.pubnub_asyncio import (
    AsyncioSubscriptionManager,
    PubNubAsyncio,
)

from .api import VivintSkyApi
from .const import (
    AuthUserAttribute,
    PubNubMessageAttribute,
    SystemAttribute,
    UserAttribute,
)
from .exceptions import VivintSkyApiError
from .pubnub import PN_CHANNEL, PN_SUBSCRIBE_KEY, VivintPubNubSubscribeListener
from .system import System
from .utils import first_or_none

_LOGGER = logging.getLogger(__name__)


class Account:
    """Class for interacting with VivintSky API using asyncio."""

    def __init__(
        self,
        username: str,
        password: str | None = None,
        refresh_token: str | None = None,
        client_session: aiohttp.ClientSession | None = None,
    ):
        """Initialize an account."""
        self.__connected = False
        self.__load_devices = False
        self.__pubnub: PubNubAsyncio | None = None
        self.__pubnub_listener: VivintPubNubSubscribeListener | None = None
        self._api = VivintSkyApi(
            username=username,
            password=password,
            refresh_token=refresh_token,
            client_session=client_session,
        )
        self.systems: list[System] = []

    @property
    def api(self) -> VivintSkyApi:
        """Return the API."""
        return self._api

    @property
    def connected(self) -> bool:
        """Return True if connected."""
        return self.__connected

    @property
    def refresh_token(self) -> str | None:
        """Return the refresh token."""
        return self.api.tokens.get("refresh_token")

    async def connect(
        self, load_devices: bool = False, subscribe_for_realtime_updates: bool = False
    ) -> None:
        """Connect to the VivintSky API."""
        _LOGGER.debug("Connecting to VivintSky")

        self.__load_devices = load_devices

        # initialize the vivintsky cloud session
        authuser_data = await self.api.connect()
        self.__connected = True

        # subscribe to pubnub for realtime updates
        if subscribe_for_realtime_updates:
            _LOGGER.debug("Subscribing to PubNub for realtime updates")
            await self.subscribe_for_realtime_updates(authuser_data)

        # load all systems, panels and devices
        if self.__load_devices:
            _LOGGER.debug("Loading devices")
            await self.refresh(authuser_data)

    async def disconnect(self) -> None:
        """Disconnect from the API."""
        _LOGGER.debug("Disconnecting from VivintSky")
        if self.connected:
            if self.__pubnub:
                self.__pubnub.remove_listener(self.__pubnub_listener)
                await self.__pubnub_unsubscribe_all()
                await self.__pubnub.stop()
        await self.api.disconnect()
        self.__connected = False

    async def __pubnub_unsubscribe_all(self) -> None:
        """
        Unsubscribe from all channels and wait for the response.

        The pubnub code doesn't properly wait for the unsubscribe event to finish or to
        be canceled, so we have to manually do it by finding the coroutine in asyncio.
        """
        assert self.__pubnub
        self.__pubnub.unsubscribe_all()
        tasks = [
            task
            for task in asyncio.all_tasks()
            if getattr(getattr(task.get_coro(), "cr_code", None), "co_name", None)
            == "_send_leave_helper"
        ]
        await asyncio.gather(*tasks)

    async def verify_mfa(self, code: str) -> None:
        """Verify multi-factor authentication with the VivintSky API."""
        await self.api.verify_mfa(code)

        # load all systems, panels and devices
        if self.__load_devices:
            _LOGGER.debug("Loading devices")
            await self.refresh()

    async def refresh(self, authuser_data: dict | None = None) -> None:
        """Refresh the account."""
        # make a call to vivint's authuser endpoint to get a list of all the system_accounts (locations) & panels if not supplied
        if not authuser_data:
            try:
                authuser_data = await self.api.get_authuser_data()
            except (ClientConnectionError, VivintSkyApiError):
                _LOGGER.error("Unable to refresh system(s)")

        if authuser_data:
            # for each system_account, make another call to load all the devices
            for system_data in authuser_data[AuthUserAttribute.USERS][
                UserAttribute.SYSTEM
            ]:
                # is this an existing account_system?
                system = first_or_none(
                    self.systems,
                    lambda system, system_data=system_data: (  # type: ignore
                        system.id == system_data[SystemAttribute.PANEL_ID]
                    ),
                )
                if system:
                    await system.refresh()
                else:
                    full_system_data = await self.api.get_system_data(
                        system_data[SystemAttribute.PANEL_ID]
                    )
                    self.systems.append(
                        System(
                            data=full_system_data,
                            api=self.api,
                            name=system_data.get(SystemAttribute.SYSTEM_NICKNAME),
                            is_admin=system_data.get(SystemAttribute.ADMIN, False),
                        )
                    )

            _LOGGER.debug(
                "Refreshed %s system(s)",
                len(authuser_data[AuthUserAttribute.USERS][UserAttribute.SYSTEM]),
            )

    async def subscribe_for_realtime_updates(
        self, authuser_data: dict | None = None
    ) -> None:
        """Subscribe to PubNub for realtime updates."""
        # make a call to vivint's authuser endpoint to get message broadcast channel if not supplied
        if not authuser_data:
            authuser_data = await self.api.get_authuser_data()

        user_data = authuser_data[AuthUserAttribute.USERS]

        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = PN_SUBSCRIBE_KEY
        pnconfig.user_id = f"pn-{user_data[UserAttribute.ID].upper()}"

        self.__pubnub = PubNubAsyncio(
            pnconfig, subscription_manager=AsyncioSubscriptionManager
        )
        self.__pubnub_listener = VivintPubNubSubscribeListener(
            self.handle_pubnub_message
        )
        self.__pubnub.add_listener(self.__pubnub_listener)

        pn_channel = (
            f"{PN_CHANNEL}#{user_data[UserAttribute.MESSAGE_BROADCAST_CHANNEL]}"
        )
        self.__pubnub.subscribe().channels(pn_channel).with_presence().execute()

    def handle_pubnub_message(self, message: dict) -> None:
        """Handle a pubnub message."""
        panel_id = message.get(PubNubMessageAttribute.PANEL_ID)
        if not panel_id:
            _LOGGER.debug(
                "PubNub message ignored (no panel id specified): %s",
                message,
            )
            return

        system = first_or_none(self.systems, lambda system: system.id == panel_id)
        if not system:
            _LOGGER.debug("No system found with id %s: %s", panel_id, message)
            return

        system.handle_pubnub_message(message)



"""Script to generate Z-Wave device data."""

from __future__ import annotations

import json
import os
from typing import Final

with open(
    os.path.join(os.path.dirname(__file__), "zjs_device_config_db.json"),
    encoding="utf8",
) as file:
    ZJS_DEVICE_DB: Final[dict[str, str | dict[str, str]]] = json.load(file)


def get_zwave_device_info(
    manufacturer_id: int | None, product_type: int | None, product_id: int | None
) -> dict[str, str]:
    """Lookup the Z-Wave device based on the manufacturer id, product type and product id."""
    key = f"0x{manufacturer_id:04x}:0x{product_type:04x}:0x{product_id:04x}"
    if isinstance((device_info := ZJS_DEVICE_DB.get(key)), dict):
        return device_info
    return {}



"""Module that implements the VivintPubNubSubscribeListener class."""

from __future__ import annotations

import logging
from collections.abc import Callable
from enum import Enum
from typing import Any

from pubnub.callbacks import SubscribeCallback
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.pubsub import PNMessageResult
from pubnub.pubnub_asyncio import PubNubAsyncio

PN_SUBSCRIBE_KEY = "sub-c-6fb03d68-6a78-11e2-ae8f-12313f022c90"
PN_CHANNEL = "PlatformChannel"

_LOGGER = logging.getLogger(__name__)


class VivintPubNubSubscribeListener(SubscribeCallback):
    """PubNub Subscribe Listener."""

    def __init__(self, message_received_callback: Callable):
        """Initialize the PubNub subscription."""
        super().__init__()
        self.__message_received_callback = message_received_callback

    def presence(self, pubnub: PubNubAsyncio, presence: Any) -> None:
        """Handle presence update."""
        _LOGGER.debug("Recieved new presence: %s", presence)

    def status(
        self,
        pubnub: PubNubAsyncio,
        status: PNStatus,
    ) -> None:
        """Handle a status update."""
        operation = op.name if isinstance(op := status.operation, Enum) else op
        category = cat.name if isinstance(cat := status.category, Enum) else cat
        if status.is_error():
            _LOGGER.error(
                "PubNub status error - operation: %s, category: %s, error: %s",
                operation,
                category,
                status.error_data.information,
            )
        else:
            _LOGGER.debug(
                "PubNub status update - operation: %s, category: %s",
                operation,
                category,
            )

    def message(self, pubnub: PubNubAsyncio, message: PNMessageResult) -> None:
        """Handle a message received."""
        self.__message_received_callback(message.message)



"""Module that implements the Thermostat class."""

from __future__ import annotations

import logging
from typing import Any

from ..const import ThermostatAttribute as Attribute
from ..enums import DeviceType, FanMode, HoldMode, OperatingMode, OperatingState
from . import VivintDevice
from .alarm_panel import AlarmPanel

_LOGGER = logging.getLogger(__name__)


class Thermostat(VivintDevice):
    """Represents a Vivint thermostat device."""

    def __init__(self, data: dict, alarm_panel: AlarmPanel):
        """Initialize a thermostat."""
        super().__init__(data, alarm_panel)

        if data.get(Attribute.ACTUAL_TYPE) == DeviceType.POD_NEST_THERMOSTAT.value:
            [self._manufacturer, self._model] = ["Google", "Nest"]

    @property
    def cool_set_point(self) -> float | None:
        """Return the cool set point of the thermostat."""
        return self.data.get(Attribute.COOL_SET_POINT)

    @property
    def fan_mode(self) -> FanMode:
        """Return the fan mode of the thermostat."""
        return FanMode(self.data.get(Attribute.FAN_MODE))  # type: ignore

    @property
    def heat_set_point(self) -> float | None:
        """Return the heat set point of the thermostat."""
        return self.data.get(Attribute.HEAT_SET_POINT)

    @property
    def hold_mode(self) -> HoldMode:
        """Return the hold mode of the thermostat."""
        return HoldMode(self.data.get(Attribute.HOLD_MODE))  # type: ignore

    @property
    def humidity(self) -> int | None:
        """Return the humidity of the thermostat."""
        return self.data.get(Attribute.HUMIDITY)

    @property
    def is_fan_on(self) -> bool:
        """Return `True` if the thermostat fan is on."""
        return self.data.get(Attribute.FAN_STATE) == 1

    @property
    def is_on(self) -> bool:
        """Return `True` if the thermostat is on."""
        return self.operating_state != OperatingState.IDLE

    @property
    def maximum_temperature(self) -> float | None:
        """Return the maximum temperature of the thermostat."""
        return self.data.get(Attribute.MAXIMUM_TEMPERATURE)

    @property
    def minimum_temperature(self) -> float | None:
        """Return the minimum temperature of the thermostat."""
        return self.data.get(Attribute.MINIMUM_TEMPERATURE)

    @property
    def operating_mode(self) -> OperatingMode:
        """Return the operating mode of the thermostat."""
        return OperatingMode(self.data.get(Attribute.OPERATING_MODE))  # type: ignore

    @property
    def operating_state(self) -> OperatingState:
        """Return the operating state of the thermostat."""
        return OperatingState(self.data.get(Attribute.OPERATING_STATE))  # type: ignore

    @property
    def temperature(self) -> float | None:
        """Return the temperature of the thermostat."""
        return self.data.get(Attribute.CURRENT_TEMPERATURE)

    @staticmethod
    def celsius_to_fahrenheit(celsius: float) -> int:
        """Convert Celsius to Fahrenheit."""
        return round(celsius * 1.8 + 32)

    async def set_state(self, **kwargs: Any) -> None:
        """Set the state of the thermostat."""
        assert self.alarm_panel
        await self.api.set_thermostat_state(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id, **kwargs
        )



"""Module that implements the AlarmPanel class."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Type

from ..api import VivintSkyApi
from ..const import AlarmPanelAttribute as Attribute
from ..const import (
    PanelUpdateAttribute,
    PubNubMessageAttribute,
    PubNubOperatorAttribute,
    SystemAttribute,
)
from ..enums import ArmedState, DeviceType
from ..exceptions import VivintSkyApiError
from ..utils import add_async_job, first_or_none
from . import VivintDevice, get_device_class

if TYPE_CHECKING:
    from ..system import System

_LOGGER = logging.getLogger(__name__)

DEVICE_DELETED = "device_deleted"
DEVICE_DISCOVERED = "device_discovered"


class AlarmPanel(VivintDevice):
    """Describe a Vivint alarm panel."""

    def __init__(self, data: dict, system: System):
        """Initialize an alarm panel."""
        self.system = system
        super().__init__(data)
        self.devices: list[VivintDevice] = []
        self.unregistered_devices: dict[int, tuple] = {}

        self.__parse_data(data=data, init=True)

        # store a reference to the physical panel device
        self.__panel_credentials: dict = {}
        self.__panel = first_or_none(
            self.devices,
            lambda device: (
                DeviceType(device.data.get(Attribute.TYPE)) == DeviceType.PANEL
            ),
        )

    def __repr__(self) -> str:
        """Return custom __repr__ of device."""
        return f"<{self.__class__.__name__} {self.id}, {self.name}: {self.state.name}>"

    @property
    def api(self) -> VivintSkyApi:
        """Return the API."""
        return self.system.api

    @property
    def id(self) -> int:
        """Panel's id."""
        return int(self.data[Attribute.PANEL_ID])

    @property
    def name(self) -> str:
        """Panel's name."""
        return self.system.name

    @property
    def mac_address(self) -> str:
        """Panel's MAC Address."""
        return str(self.data[Attribute.MAC_ADDRESS])

    @property
    def manufacturer(self) -> str:
        """Return Vivint as the manufacturer of this panel."""
        return "Vivint"

    @property
    def model(self) -> str:
        """Return the model of the physical panel."""
        return (
            "Sky Control"
            if self.__panel and self.__panel.data["pant"] == 1
            else "Smart Hub"
        )

    @property
    def software_version(self) -> str | None:
        """Return the software version of the panel."""
        return self.__panel.software_version if self.__panel else None

    @property
    def partition_id(self) -> int:
        """Panel's partition id."""
        return int(self.data[Attribute.PARTITION_ID])

    @property
    def is_disarmed(self) -> bool:
        """Return True if alarm is disarmed."""
        return self.state == ArmedState.DISARMED

    @property
    def is_armed_away(self) -> bool:
        """Return True if alarm is in armed away state."""
        return self.state == ArmedState.ARMED_AWAY

    @property
    def is_armed_stay(self) -> bool:
        """Return True if alarm is in armed stay state."""
        return self.state == ArmedState.ARMED_STAY

    @property
    def state(self) -> ArmedState:
        """Return the panel's armed state."""
        return ArmedState(self.data.get(Attribute.STATE))  # type: ignore

    @property
    def credentials(self) -> dict:
        """Return the panel credentials."""
        return self.__panel_credentials

    async def set_armed_state(self, state: int) -> None:
        """Set the armed state for a panel."""
        _LOGGER.debug("Setting %s to %s", self.name, ArmedState(state).name)
        await self.api.set_alarm_state(self.id, self.partition_id, state)

    async def trigger_alarm(self) -> None:
        """Trigger an alarm."""
        _LOGGER.debug("Triggering an alarm on %s", self.name)
        await self.api.trigger_alarm(self.id, self.partition_id)

    async def disarm(self) -> None:
        """Disarm the alarm."""
        await self.set_armed_state(ArmedState.DISARMED)

    async def arm_stay(self) -> None:
        """Set the alarm to armed stay."""
        await self.set_armed_state(ArmedState.ARMED_STAY)

    async def arm_away(self) -> None:
        """Set the alarm to armed away."""
        await self.set_armed_state(ArmedState.ARMED_AWAY)

    async def get_panel_credentials(self, refresh: bool = False) -> dict:
        """Get the panel credentials."""
        if refresh or not self.__panel_credentials:
            self.__panel_credentials = await self.api.get_panel_credentials(self.id)
        return self.__panel_credentials

    async def get_software_update_details(self) -> dict[str, bool | str]:
        """Get the software update details."""
        if not self.system.is_admin:
            _LOGGER.warning(
                "%s - Cannot get software update details as user is not an admin",
                self.name,
            )
            details = {}
        else:
            details = await self.api.get_system_update(self.id)
        return {
            "available": details.get(PanelUpdateAttribute.AVAILABLE, False),
            "available_version": details.get(
                PanelUpdateAttribute.AVAILABLE_VERSION, ""
            ),
            "current_version": details.get(PanelUpdateAttribute.CURRENT_VERSION, ""),
            "update_reason": details.get(PanelUpdateAttribute.UPDATE_REASON, ""),
        }

    async def update_software(self) -> bool:
        """Update the panel software version."""
        if not self.system.is_admin:
            _LOGGER.warning(
                "%s - Cannot update software as user is not an admin", self.name
            )
            return False
        try:
            await self.api.update_panel_software(self.id)
        except VivintSkyApiError as err:
            _LOGGER.error("%s - %s", self.name, err)
            return False
        return True

    async def reboot(self) -> None:
        """Reboot the panel."""
        if not self.system.is_admin:
            _LOGGER.warning(
                "%s - Cannot reboot panel as user is not an admin", self.name
            )
            return
        await self.api.reboot_panel(self.id)

    def get_devices(
        self, device_types: set[Type[VivintDevice]] | None = None
    ) -> list[VivintDevice]:
        """Get a list of associated devices."""
        devices: list[VivintDevice] = []

        if device_types:
            devices = [
                device for device in self.devices if device.__class__ in device_types
            ]
        else:
            devices = self.devices

        return devices

    def refresh(self, data: dict, new_device: bool = False) -> None:
        """Refresh the alarm panel."""
        if not new_device:
            self.update_data(data, override=True)
        else:
            self.data[Attribute.DEVICES].extend(data[Attribute.DEVICES])

        self.__parse_data(data)

    def handle_pubnub_message(self, message: dict) -> None:
        """Handle a pubnub message."""
        operation = message.get(PubNubMessageAttribute.OPERATION)
        data = message.get(PubNubMessageAttribute.DATA)
        if not data:
            _LOGGER.debug(
                "Ignoring account partition message for panel %s, partition %s (no data provided): %s",
                self.id,
                self.partition_id,
                message,
            )
            return

        if not data.get(PubNubMessageAttribute.DEVICES):
            # this message is for the panel itself
            self.update_data(data)
        else:
            # this is a message for one (or more) of the attached devices
            devices_data = data[PubNubMessageAttribute.DEVICES]

            for device_data in devices_data:
                device_id = device_data.get("_id")
                if not device_id:
                    _LOGGER.debug("No device id")
                    continue

                if operation == PubNubOperatorAttribute.CREATE:
                    self.refresh(data=data, new_device=True)
                    add_async_job(self.handle_new_device, device_id)
                else:
                    device = first_or_none(
                        self.devices,
                        lambda device, device_id=device_id: device.id == device_id,  # type: ignore
                    )
                    if not device:
                        _LOGGER.debug(
                            "Ignoring message for device %s (device not found)",
                            device_id,
                        )
                        continue

                    # for the sake of consistency, we also need to update the panel's raw data
                    raw_device_data = first_or_none(
                        self.data[Attribute.DEVICES],
                        lambda raw_device_data,  # type: ignore
                        device_data=device_data: (
                            raw_device_data["_id"] == device_data["_id"]
                        ),
                    )

                    if operation == PubNubOperatorAttribute.DELETE:
                        self.devices.remove(device)
                        if raw_device_data is not None:
                            self.data[Attribute.DEVICES].remove(raw_device_data)
                        self.unregistered_devices[device.id] = (
                            device.name,
                            device.device_type,
                        )
                        self.emit(DEVICE_DELETED, {"device": device})
                    else:
                        device.handle_pubnub_message(device_data)
                        assert raw_device_data
                        raw_device_data.update(device_data)

    async def handle_new_device(self, device_id: int) -> None:
        """Handle a new device."""
        try:
            device = first_or_none(self.devices, lambda device: device.id == device_id)
            assert device
            while not device.is_valid:
                await asyncio.sleep(1)
                if device.id in self.unregistered_devices:
                    return
            resp = await self.api.get_device_data(self.id, device_id)
            data = resp[SystemAttribute.SYSTEM][SystemAttribute.PARTITION][0]
            self.refresh(data, new_device=True)
            self.emit(DEVICE_DISCOVERED, {"device": device})
        except VivintSkyApiError:
            _LOGGER.error("Error getting new device data for device %s", device_id)

    def __parse_data(self, data: dict, init: bool = False) -> None:
        """Parse the alarm panel data."""
        for device_data in data[Attribute.DEVICES]:
            device: VivintDevice | None = None
            if not init:
                device = first_or_none(
                    self.devices,
                    lambda device, device_data=device_data: (  # type: ignore
                        device.id == device_data[Attribute.ID]
                    ),
                )
            if device:
                device.update_data(device_data, override=True)
            else:
                self.__parse_device_data(device_data=device_data)

        if data.get(Attribute.UNREGISTERED):
            self.unregistered_devices = {
                device[Attribute.ID]: (
                    device[Attribute.NAME],
                    DeviceType(device[Attribute.TYPE]),
                )
                for device in data[Attribute.UNREGISTERED]
            }

    def __parse_device_data(self, device_data: dict) -> None:
        """Parse device data and optionally emit a device discovered event."""
        device_class = get_device_class(device_data[Attribute.TYPE])
        device = device_class(device_data, self)
        self.devices.append(device)
